var timerID;
var startTime, zeroHour, remainingTime;
var keyBuffer;
var rewritingMode;
var repeatMode;
var display = "front";
var sound, keyEvents, timerGraph, notification;
var startButton, setButtonA, setButtonB;
var inc43Button, inc21Button, dec43Button, dec21Button;

// Alarm sound file and repeat times (0 to 65535)
var soundFile = "/System/Library/Sounds/Purr.aiff";
var soundRepeat = 10;
var soundRepeatShort = 5;

// Skin Files
var jacketFiles = [ "Gr", "Yl", "Bl", "Or", "Wt", "Rd", "Bk", "Pk", "Br", "Gl" ];
var jacketSampleContext;
var widgetPref;
var prefPane;
var websiteURI = "http://www.cyanworks.net/dboardKTimer.html"

// Widget Initialize
function init() {
	timerID = null;
	zeroHour = null;
	startTime = null;
	remainingTime = 0;
	rewritingMode = true;
	keyBuffer = null;

	widgetPref = new WidgetPreference();
	widgetPref.jacketsNum = jacketFiles.length;
	if (window.widget) {
		widget.onshow = onShow;
		widget.onhide = onHide;
		widget.onremove = onRemove;
	}

	// Sound
	sound = new Sound();
	sound.load(soundFile);

	initControlButtons();

	timerGraph = new TimerGraph();
	notification = new Notification();

	// Load preferences
	widgetPref.load();
	setTimer(widgetPref.value("timer"));
	setJacket(widgetPref.value("jacket"));
	setRepeatMode(widgetPref.value("repeatMode"));
    
    // Balloon
    rewritePresetBalloon("timerA", widgetPref.value("timerA"));
    rewritePresetBalloon("timerB", widgetPref.value("timerB"));

	// Prepare the preference section
	prefPane = new PreferencePane(widgetPref);

	// EventHandlers
	keyEvents = new KeyEvents();
}

// Create buttons
function initControlButtons() {
	startButton = new Button(document.getElementById("startbutton"),
	  "Images/startButton0.png", "Images/startButton1.png", null);
	setButtonA  = new Button(document.getElementById("setbuttonA"),
	  "Images/setButton0.png",  "Images/setButton1.png",  null);
	setButtonB  = new Button(document.getElementById("setbuttonB"),
	  "Images/setButton0.png",  "Images/setButton1.png",  null);
	inc43Button = new Button(document.getElementById("inc43button"),
	  "Images/inc43Button0.png", "Images/inc43Button1.png", null);
	inc21Button = new Button(document.getElementById("inc21button"),
	  "Images/inc21Button0.png", "Images/inc21Button1.png", null);
	dec43Button = new Button(document.getElementById("dec43button"),
	  "Images/dec43Button0.png", "Images/dec43Button1.png", null);
	dec21Button = new Button(document.getElementById("dec21button"),
	  "Images/dec21Button0.png", "Images/dec21Button1.png", null);

	// Properties of Start/Stop button
	startButton.action      = function() { switchTimer(); };
	startButton.setRegion("control circle 0 0 0 0");
	startButton.checkPosition = function(x, y) {
		return(Math.pow((x - 21) * 14, 2) + Math.pow((14 - y) * 21, 2) <= Math.pow(14 * 21, 2));
	};

	// Properties of Preset button A
	setButtonA.action       = function() { loadPreset("timerA"); };
	setButtonA.pressAction  = function() { savePreset("timerA"); };
	setButtonA.overAction   = function() {
        var balloon = document.getElementById("balloonpreseta");
        balloon.style.visibility = "visible";
        balloon.style.zIndex = 5;
    };
	setButtonA.outAction   = function() {
        var balloon = document.getElementById("balloonpreseta");
        balloon.style.visibility = "hidden";
        balloon.style.zIndex = 0;
    };
	setButtonA.setRegion("control circle 0 0 0 0");
	setButtonA.checkPosition = function(x, y) {
		return(Math.pow(x - 10, 2) + Math.pow(10 - y, 2) <= Math.pow(10, 2));
	};

	// Properties of Preset button B
	setButtonB.action       = function() { loadPreset("timerB"); };
	setButtonB.pressAction  = function() { savePreset("timerB"); };
	setButtonB.overAction   = function() {
        var balloon = document.getElementById("balloonpresetb");
        balloon.style.visibility = "visible";
        balloon.style.zIndex = 5;
    };
	setButtonB.outAction   = function() {
        var balloon = document.getElementById("balloonpresetb");
        balloon.style.visibility = "hidden";
        balloon.style.zIndex = 0;
    };
	setButtonB.setRegion("control circle 0 0 0 0");
	setButtonB.checkPosition = function(x, y) {
		return(Math.pow(x - 10, 2) + Math.pow(10 - y, 2) <= Math.pow(10, 2));
	};

	// Properties of Increment button 43
	inc43Button.action      = function() { setTimerDelta( 1 * 60 * 1000); };
	inc43Button.shiftAction = function() { setTimerDelta(10 * 60 * 1000); };
	inc43Button.pressAction      = inc43Button.action;
	inc43Button.pressShiftAction = inc43Button.shiftAction;
	inc43Button.pressRepeatInterval = 0.05 * 1000;
	inc43Button.setRegion("control circle 0 -43 -43 0");
	inc43Button.checkPosition = function(x, y) {
		return(Math.pow(x - 43, 2) + Math.pow(43 - y, 2) <= Math.pow(42, 2));
	};

	// Properties of Increment button 21
	inc21Button.action      = function() { setTimerDelta( 1 * 1000); };
	inc21Button.shiftAction = function() { setTimerDelta(10 * 1000); };
	inc21Button.pressAction      = inc21Button.action;
	inc21Button.pressShiftAction = inc21Button.shiftAction;
	inc21Button.pressRepeatInterval = 0.05 * 1000;
	inc21Button.setRegion("control circle 0 0 -43 -43");
	inc21Button.checkPosition = function(x, y) {
		return(Math.pow(x, 2) + Math.pow(43 - y, 2) <= Math.pow(42, 2));
	};

	// Properties of Decrement button 43
	dec43Button.action      = function() { setTimerDelta( -1 * 60 * 1000); };
	dec43Button.shiftAction = function() { setTimerDelta(-10 * 60 * 1000); };
	dec43Button.pressAction      = dec43Button.action;
	dec43Button.pressShiftAction = dec43Button.shiftAction;
	dec43Button.pressRepeatInterval = 0.05 * 1000;
	dec43Button.setRegion("control circle -43 -43 0 0");
	dec43Button.checkPosition = function(x, y) {
		return(Math.pow(x - 43, 2) + Math.pow(0 - y, 2) <= Math.pow(42, 2));
	};

	// Properties of Decrement button 21
	dec21Button.action      = function() { setTimerDelta( -1 * 1000); };
	dec21Button.shiftAction = function() { setTimerDelta(-10 * 1000); };
	dec21Button.pressAction      = dec21Button.action;
	dec21Button.pressShiftAction = dec21Button.shiftAction;
	dec21Button.pressRepeatInterval = 0.05 * 1000;
	dec21Button.setRegion("control circle -43 0 0 -43");
	dec21Button.checkPosition = function(x, y) {
		return(Math.pow(x, 2) + Math.pow(0 - y, 2) <= Math.pow(42, 2));
	};

}

function onShow() {
	rewritingMode = true;
	runInterval();
}

function onHide() {
	rewritingMode = false;

	var saveWait = (1 * 1000 + Math.round(Math.random() * 1 * 1000)); // Wait 1 - 2 sec
	if (startTime != null)
		setTimeout(widgetPref.setValue(startTime, "timer"), saveWait);
}

function onRemove() {
	widgetPref.removeMyPref();
}

// Waked up from the sleep mode
function systemDidWake() {
	LOG__("systemDidWake");
	runInterval();
}

// Work Timer
function switchTimer() {
	// When notification is active
	if (notification.flag) {
		notification.end();
		if (0 < startTime)
			setTimer(startTime);
		return;
	}

	if (timerID != null)
		timerStop();   // if Timer working, turn off.
	else
		timerStart();  // if Timer sleeping, turn on.
}

function timerStart() {
	var startHour = (new Date).getTime();

	notification.end();
	keyBuffer = null;

	if ((0 < remainingTime ) && (remainingTime <= 99*60*1000 + 59*1000)) {
		zeroHour = startHour + remainingTime;
		runInterval();
	}
	rewrite();
}

function timerStop() {
	notification.end();

	if (timerID) {
		clearTimeout(timerID);
		timerID = null;

		// When remaining time is not zero, record it.
		if (zeroHour != null) {
			remainingTime = zeroHour - (new Date).getTime();
			zeroHour = null;
		}
	}

	rewrite();
}


function runInterval() {
	if (zeroHour == null)
		return;

	var interval;
	LOG2__("runInterval remains:" + remainingTime);

	remainingTime = zeroHour - (new Date).getTime();
	if (remainingTime <= 0) {  // It's Time!
        if (repeatMode) {  // Repeat the countdown
            while (remainingTime <= 0) {
                zeroHour = zeroHour + startTime;
                remainingTime = zeroHour - (new Date).getTime();
            }
            setTimer(remainingTime, startTime);
            timerStart();
        } else {
            remainingTime = 0;
            zeroHour = null;
            timerStop();
        }
		notification.start();
		return;
	} else {
		if (rewritingMode)
			rewrite();
	}

	// Pace maker
	clearTimeout(timerID);
	remainingTime = zeroHour - (new Date).getTime();
	if (rewritingMode || (remainingTime <= 10*1000))
		interval = (remainingTime % 1000 <= 0)? 1.0*1000 : remainingTime % 1000;
	else
		interval = 5.0*1000;  // idle

	timerID = setTimeout("runInterval()", interval);
}

// ================================================
// Notification
// ================================================
function Notification() {
	LOG__("Object: Notification");

	this.flag = false;
	this.badge = document.getElementById("notification");
}

Notification.prototype.start = function() {
	LOG2__("Notification.start()");

	this.flag = true;
	this.badge.style.display = "block";
	sound.play();
}

Notification.prototype.end = function() {
	LOG2__("Notification.end()");

	this.flag = false;
	this.badge.style.display = "none";
	sound.stop();
}

// ================================================
// TimerGraph
// ================================================
function TimerGraph() {
	LOG__("Object: TimerGraph");

	this.timerCanvas = document.getElementById("graph");
	this.value = 0;

	this.ringImg = new Image();
	this.ringImg.src = "Images/ring.png";

	// Events
	var localThis = this;
	this.drawImgLater = function() { localThis.drawImg(); };
}

TimerGraph.prototype.setValue = function(value) {
	this.value = value;
	this.drawImg();
}

TimerGraph.prototype.drawImg = function() {
	var ctxt = this.timerCanvas.getContext("2d");
	var angle = Math.PI * 2 * 300 / 360 * this.value;
	var angleOffset = Math.PI * 2 * 120 / 360;
	
	// If images are not loaded
	if (!(this.ringImg.complete))
		setTimeout(this.drawImgLater, 100);

	ctxt.setFillColor(0.0, 0.0, 0.5, 1.0);
	ctxt.fillRect(0, 0, 84, 84);
	ctxt.clearRect(0, 0, 84, 84);

	// Move to center
	ctxt.save();
	ctxt.translate(84/2, 84/2)

	// Draw a mask
	ctxt.save();
	ctxt.lineWidth = 6;
	ctxt.strokeStyle = "rgba(0, 0, 0, 1.0)";
	ctxt.arc(0, 0, 38, angleOffset, angleOffset + angle, 0);
	ctxt.stroke();
	ctxt.restore();

	// Draw the image in a mask
	ctxt.globalCompositeOperation = "source-in";
	ctxt.drawImage(this.ringImg, -84/2, -84/2);
	ctxt.globalCompositeOperation = "source-over";

	ctxt.restore();
}

// Rewrite Clock
function rewrite() {
	var timer = document.getElementById("timer");
	var timerMin = document.getElementById("timermin");
	var timerSec = document.getElementById("timersec");
	var graph = document.getElementById("graph");
	var min, sec;
	var digit43, digit21;

	if (keyBuffer == null) {
		// Display remaining time
		min = Math.floor(remainingTime/1000/60);
		sec = Math.ceil((remainingTime - min*60*1000)/1000);
		if (60 <= sec) {
			min++;
			sec = 0;
		}
		timerMin.innerHTML = (min < 10)? ("0" + min) : min;
		timerSec.innerHTML = (sec < 10)? ("0" + sec) : sec;
		timerGraph.setValue(remainingTime / startTime);
	} else {
		// Display numbers from keyboard
		digit43 = Math.floor((keyBuffer - Math.floor(keyBuffer/10000)*10000)/100);
		digit21 = keyBuffer - Math.floor(keyBuffer/100)*100;
		timerMin.innerHTML = (digit43 < 10)? ("0" + digit43) : digit43;
		timerSec.innerHTML = (digit21 < 10)? ("0" + digit21) : digit21;
		timerGraph.setValue(1.0);
	}

	// Change Clock Style
	if (timerID != null) {
		timer.style.color = "#ffffff";
		graph.style.opacity = 1.0;
	} else {
		timer.style.color = (keyBuffer == null)? "#999999" : "#667766";
		graph.style.opacity = 0.4;
	}
}

// Timer Set
function setTimer(renainingTimeReq, startTimeReq) {
	var request;
    timerStop();
    
    // Set Remainig Time
    request = parseInt(renainingTimeReq);
	if (request == null)
		request = 0;
	if (request < 0)
		request = 0;
	if (100*60*1000 <= request) 
		request = 100*60*1000 - 1*1000;
	remainingTime = request;

    // Set Start Time
    request = parseInt(startTimeReq);
	if (startTimeReq == null)
		request = remainingTime;
	if (request < 0)
		request = 0;
	if (100*60*1000 <= request) 
		request = 100*60*1000 - 1*1000;
	startTime = request;

	rewrite();
}

function setTimerDelta(str) {
	var delta = parseInt(str);
	var request;

	if (delta == null)
		return;
	request = Math.ceil(remainingTime/1000)*1000 + delta;
	setTimer(request);
}

function timerReset() {
	notification.end();
	setTimer(0);
}

// Preset Buttons
function savePreset(key) {
	widgetPref.setValue(Math.ceil(remainingTime/1000)*1000, key);
	keyBuffer = null;
    rewritePresetBalloon(key, remainingTime);
	rewrite();
	sound.beep();
}

function loadPreset(key) {
	timerStop();
	keyBuffer = null;
	setTimer(widgetPref.value(key));
}

// Update Balloon
function rewritePresetBalloon(key, presetTime) {
	var balloonA = document.getElementById("balloonpreseta");
	var balloonB = document.getElementById("balloonpresetb");
    var balloonText;
    var min, sec;

    min = Math.floor(presetTime/1000/60);
    sec = Math.ceil((presetTime - min*60*1000)/1000);
    if (60 <= sec) {
        min++;
        sec = 0;
    }
    balloonText = (min < 10)? ("0" + min) : min;
    balloonText += ":";
    balloonText += (sec < 10)? ("0" + sec) : sec;

    switch(key) {
        case "timerA":
            balloonA.innerHTML = balloonText;
            break;
        case "timerB":
            balloonB.innerHTML = balloonText;
            break;
    }
}

// Append a number to keyBuffer
function numImput(num) {
	var keyBufferNow;
	var keyBufferNext;
	var digit43;
	var digit21;

	if (keyBuffer == null)
		keyBuffer = 0;

	keyBufferNext = keyBuffer*10 + num;
	keyBuffer = keyBufferNext - (Math.floor(keyBufferNext/10000))*10000;

	digit43 = Math.floor(keyBuffer/100);
	digit21 = keyBuffer - Math.floor(keyBuffer/100)*100;

    keyBufferNow = keyBuffer;
	setTimer((digit43*60 + digit21)*1000);
    keyBuffer = keyBufferNow;
}

// Repeat Button
function switchRepeatMode(rev) {
	if (!repeatMode)
        setRepeatMode(true);
    else
        setRepeatMode(false);
}

function setRepeatMode(mode) {
	var repeatbutton = document.getElementById("repeatbutton");
    repeatMode = mode;
    if (mode)
        repeatbutton.src = "Images/repeatOn.png";
    else
        repeatbutton.src = "Images/repeatOff.png";

	widgetPref.setValue(mode, "repeatMode");
}

// Change jacket
function changeJacket(rev) {
	var newJacket;
	if (!rev) {
		newJacket = widgetPref.value("jacket") + 1;
		if (jacketFiles.length <= newJacket )
			newJacket = 0;
	} else {
		newJacket = widgetPref.value("jacket") - 1;
		if (newJacket < 0)
			newJacket = jacketFiles.length - 1;
	}
	setJacket(newJacket);
}

function setJacket(num) {
    var colorCode;
	var infobutton = document.getElementById("infobutton").object;

	widgetPref.setValue(num, "jacket");
    colorCode = jacketFiles[widgetPref.value("jacket")];

    // Set Images
	document.getElementById("front").style.backgroundImage
		= 'url("Images/jacket' + colorCode + '.png")';
    AlarmWidgetPlugin.setGrowlIcon("Images/Icon" + colorCode + ".png");

    /*
    startButton.setImageDefault("Images/startButton0" + colorCode + ".png");
    startButton.setImageClicked("Images/startButton1" + colorCode + ".png");
    setButtonA.setImageDefault("Images/setButton0" + colorCode + ".png");
    setButtonA.setImageClicked("Images/setButton1" + colorCode + ".png");
    setButtonB.setImageDefault("Images/setButton0" + colorCode + ".png");
    setButtonB.setImageClicked("Images/setButton1" + colorCode + ".png");
    */

	switch (colorCode) {
		case "Gr":
		case "Gl":
		case "Yl":
		case "Wt":
		case "Pk":
			infobutton.setStyle("black", "black");
			break;
		default: 
			infobutton.setStyle("white", "white");
			break;
	}
}

// Open website
function openWebsite() {
	if (window.widget)
		widget.openURL(websiteURI);
	else
		window.location.href = websiteURI;
}

// Reverse sides and preferences
function showBack(event) {
	if (window.widget)
		widget.prepareForTransition("ToBack");

	prefPane.setValue(widgetPref.value("jacket"), "jacket");

	document.getElementById("front").style.display = "none";
	document.getElementById("back").style.display = "block";
	display = "back";
	
	if (window.widget)
		setTimeout ("widget.performTransition();", 0);  
}

function showFront(event) {
	if (window.widget)
		widget.prepareForTransition("ToFront");

	document.getElementById("front").style.display = "block";
	document.getElementById("back").style.display = "none";
	display = "front";

	prefPane.apply();
	setTimeout ("rewrite();", 100);

	if (window.widget)
		setTimeout ("widget.performTransition();", 0);
}


// ================================================
// EventHandler (keyboard)
// ================================================
function KeyEvents() {
	LOG__("Object: KeyEvents");
	document.addEventListener("keydown",  this.onKeyDown, true);
	document.addEventListener("keypress", this.onKeyPress, true);
	document.addEventListener("keyup",    this.onKeyUp, true);
}

KeyEvents.prototype.onKeyDown = function(theEvent) {
	LOG2__("onKeyDown charCode:" + theEvent.charCode);
	var handled = true;

	if (display == "front") {
		switch(theEvent.charCode) {
			case 63232:  // up
				if (!theEvent.shiftKey)
					inc43Button.keyDown();
				else
					inc43Button.keyDownShift();
				break;
			case 63233:  // down
				if (!theEvent.shiftKey)
					dec43Button.keyDown();
				else
					dec43Button.keyDownShift();
				break;
			case 63235: // right
				if (!theEvent.shiftKey)
					inc21Button.keyDown();
				else
					inc21Button.keyDownShift();
				break;
			case 63234:  // left
				if (!theEvent.shiftKey)
					dec21Button.keyDown();
				else
					dec21Button.keyDownShift();
				break;
			case 32:  // Space bar
				startButton.keyDown();
				break;
			case 8:      // delete
			case 63272:  // del
			case 63289:  // clear
				timerReset();
				break;
			case 116:  // t
			case 84:   // T
				sound.test();
				break;
			case 119:  // w
			case 87:   // W
				openWebsite();
				break;
			case 97:  // a
			case 65:  // A
				setButtonA.keyDown();
				break;
			case 98:  // b
			case 66:  // B
				setButtonB.keyDown();
				break;
			case 97:   // a
			case 65:   // A
			case 98:   // b
			case 66:   // B
			case 115:  // s
			case 83:   // S
			case 9:    // Tab
				break;

			// numbaer input 0 - 9
			case 48:
			case 49:
			case 50:
			case 51:
			case 52:
			case 53:
			case 54:
			case 55:
			case 56:
			case 57:
				numImput(theEvent.charCode - 48);
				break;
			case 13:  // return
			case 3:   // enter
				if (keyBuffer != null) {
					keyBuffer = null;
					rewrite();
				}
				break;
			default:
				handled = false;
				break;
		}
	}

	if (handled) {
		theEvent.stopPropagation();
		theEvent.preventDefault();
	}
}

KeyEvents.prototype.onKeyPress = function(theEvent) {
	LOG2__("onKeyPress charCode:" + theEvent.charCode);
	var handled = true;

	if (display == "front") {
		switch(theEvent.charCode) {
			case 63232:  // up
			case 63233:  // down
			case 63235:  // right
			case 63234:  // left
				break;
			case 63276:  // pageup
				setTimer(remainingTime + 10 * 60 * 1000);
				break;
			case 63277:  // pagedown
				setTimer(remainingTime - 10 * 60 * 1000);
				break;
			case 32:  // space
			case 97:  // a
			case 65:  // A
			case 98:  // b
			case 66:  // B
				break;
			default:
				handled = false;
				break;
		}
	}

	if (handled) {
		theEvent.stopPropagation();
		theEvent.preventDefault();
	}
}

KeyEvents.prototype.onKeyUp = function(theEvent) {
	var handled = true;
	LOG2__("onKeyUp charCode:" + theEvent.charCode);

	if (display == "front") {
		switch(theEvent.charCode) {
			case 63232:  // Up
				if (!theEvent.shiftKey)
					inc43Button.keyUp();
				else
					inc43Button.keyUpShift();
				break;
			case 63233:  // Down
				if (!theEvent.shiftKey)
					dec43Button.keyUp();
				else
					dec43Button.keyUpShift();
				break;
			case 63235: // right
				if (!theEvent.shiftKey)
					inc21Button.keyUp();
				else
					inc21Button.keyUpShift();
				break;
			case 63234:  // left
				if (!theEvent.shiftKey)
					dec21Button.keyUp();
				else
					dec21Button.keyUpShift();
				break;
			case 32:  // Space bar
				startButton.keyUp();
				break;
			case 97:  // a
			case 65:  // A
				setButtonA.keyUp();
				break;
			case 98:  // b
			case 66:  // B
				setButtonB.keyUp();
				break;
			case 114:  // r
			case 82:   // R
                switchRepeatMode();
				break;
			case 115:  // s
			case 83:   // S
			case 9:    // Tab
				if (!theEvent.shiftKey)
					changeJacket(false);
				else
					changeJacket(true);
				break;
			default:
				handled = false;
				break;
		}
	}

	if (handled) {
		theEvent.stopPropagation();
		theEvent.preventDefault();
	}
}

// ================================================
// Sound
// ================================================
function Sound() {
	LOG__("Object: Sound");
	if (!(window.widget && AlarmWidgetPlugin))
		LOG__("Cannot load the Widget plugin");
}

// Load the sound file
Sound.prototype.load = function(uri) {
	if (window.widget && AlarmWidgetPlugin)
		AlarmWidgetPlugin.load(uri);
	else
		LOG__("Widget plugin not loaded: loadSound failed.");
}

// Stop playing sound
Sound.prototype.stop = function() {
	if (window.widget && AlarmWidgetPlugin)
		AlarmWidgetPlugin.stop();
}

// Play the sound
Sound.prototype.play = function() {
	if (window.widget && AlarmWidgetPlugin)
        if (!repeatMode)
            AlarmWidgetPlugin.play(soundRepeat);
        else
            AlarmWidgetPlugin.play(soundRepeatShort);
	else
		alert("It's Time!");
}

// Test the sound
Sound.prototype.test = function() {
	if (window.widget && AlarmWidgetPlugin)
		AlarmWidgetPlugin.play(1);
}

// Test the beep
Sound.prototype.beep = function() {
	if (window.widget && AlarmWidgetPlugin)
		AlarmWidgetPlugin.beep();
}

// ================================================
// Debugging
// ================================================
function LOG__(arg) {
	return;
	if (window.widget)
		alert(arg);
	else
		if (window.console) window.console.log(arg);
}

function LOG2__(arg) {
	return;
	LOG__(arg);
}

function LOG3__(arg) {
	return;
	LOG__(arg);
}
